//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dron - Arduino
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Impementación de un UAV tipo cuadricóptero con motores Brushless (sin escobillas)
// El control se realiza usando la velocidad angular y haciendo correcciones según los ángulos calculados
// El control remoto está compuesto por transmisor y receptor
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Incluimos algunas librerías adicionales que vamos a utilizar
#include <Wire.h>               // Librería para protocolo I2C
#include<Servo.h>               // Librería para el manejo de los motores
#include <avr/wdt.h>            // Librería para el Watchdog

// ALGUNAS DEFINICIONES
//Direccion I2C de la IMU
#define IMU_ADDRESS 0x68

// Pines para el LED y el nivel de la batería
#define LED_PIN 2
#define BATTERY A5

// Variables control PID
unsigned long lastTime;
double Input_r, Output_r, Setpoint_r;
double Input_p, Output_p, Setpoint_p;
double Input_y, Output_y, Setpoint_y;
double Output_h;
double ITerm_r, lastInput_r;
double ITerm_p, lastInput_p;
double ITerm_y, lastInput_y;
double kp_r, ki_r, kd_r;
double kp_p, ki_p, kd_p;
double kp_y, ki_y, kd_y;
int SampleTime = 4;                     // 4 milisegundos
double outMin_dif, outMax_dif;
float roll_level_adjust, pitch_level_adjust;

// Offset para los acelerómetros (deben ser obtenidos en reposo en posición horizontal - usar el programa de la IMU)
float aoffsetX = -8, aoffsetY = -199, aoffsetZ = 3870;

// Variables IMU
int acc[3];
int gyro[3];
int temperature;
float rawX, rawY, rawZ;
float X, Y, Z;
float gyroX, gyroY, gyroZ;
float  gx_rate, gy_rate, gz_rate;
float  last_gx_rate, last_gy_rate, last_gz_rate;
float rollrad, pitchrad;
unsigned long mtime, looptime, loop_timer;
float angleAx, angleAy, angleAz;
float angleGx, angleGy, angleGz;
float goffsetX, goffsetY, goffsetZ;

// Variables del filtro
float angleFx, angleFy;

// Motores
Servo esc1, esc2, esc3, esc4;
char dato;

// Control remoto
volatile int rollref, pitchref, yawref, href;
byte last_channel_1, last_channel_2, last_channel_3, last_channel_4;
int receiver_input[4];
unsigned long timer_1, timer_2, timer_3, timer_4, current_time, last_int;

// Otras
int start;
int led_dis;
float battery_voltage;


void setup()
{
  // Deshabilitar watchdog antes del setup
  wdt_disable();
  // Inicio del puerto serial a 115200 baudios y mensaje de bienvenida
  Serial.begin(115200);  
  Serial.println("Bienvenido, sistema iniciando...");
  // Inicio de comunicación I2C
  Wire.begin();
  // Inicializamos la IMU
  init_IMU();  

  // Asignamos cada motor a su pin correspondiente
  esc1.attach(8,1000,2000);
  esc2.attach(9,1000,2000);
  esc3.attach(10,1000,2000);
  esc4.attach(11,1000,2000);
  esc1.write(0);              // ¡IMPORTANTE! Hay que armar cada motor
  esc2.write(0);
  esc3.write(0);
  esc4.write(0);

  // Los pines que se utilicen como salida deben declararse explícitamente - en este caso, el LED
  pinMode(LED_PIN, OUTPUT);  

  // Un pequeño "delay" e iniciamos la calibración de la IMU
  delay(1000);
  Serial.println("Calibrando...");
  Serial.println("Mantener la IMU quieta y en posición horizontal");
  calibrate();
  delay(500);
  Serial.println("Calibración finalizada.");

  // Inicialización de los parámetros de control
  double *limsdif = SetOutputLimits(-400, 400);
  outMin_dif = limsdif[0];
  outMax_dif = limsdif[1];
  ITerm_r = 0, lastInput_r = 0;
  ITerm_p = 0, lastInput_p = 0;
  ITerm_y = 0, lastInput_y = 0;
  // Los controles se asumen centrados al comienzo
  rollref = 1500;
  pitchref = 1500;
  yawref = 1500;
  href = 0;                     // Lo iniciamos en cero como valor seguro.
  Output_r = 0, Output_p = 0; Output_y = 0;
  // Roll Control
  double *cons_r = SetTunings(0.8, 0.1, 0.02);
  kp_r = cons_r[0];
  ki_r = cons_r[1];
  kd_r = cons_r[2];
  // Pitch Control
  // Igualamos a los parámetros de "roll" ya que el dron es prácticamente simétrico en los ejes x y y
  double *cons_p = cons_r;
  kp_p = cons_p[0];
  ki_p = cons_p[1];
  kd_p = cons_p[2];
  // Yaw Control  
  double *cons_y = SetTunings(2, 0.01, 0.01);
  kp_y = cons_y[0];
  ki_y = cons_y[1];
  kd_y = cons_y[2];  

  // Configuramos las interrupciones para recibir los comandos remotos
  PCICR |= (1 << PCIE2);                                                     // PCIE2 activado para habilitar PCMSK2 scan.
  PCMSK2 |= (1 << PCINT23);                                                  // PCINT23 (D7) activado para disparar una interrupción si hay cambio de estado.
  PCMSK2 |= (1 << PCINT22);                                                  // PCINT22 (D6) activado para disparar una interrupción si hay cambio de estado.
  PCMSK2 |= (1 << PCINT21);                                                  // PCINT21 (D5) activado para disparar una interrupción si hay cambio de estado.
  PCMSK2 |= (1 << PCINT20);                                                  // PCINT20 (D4) activado para disparar una interrupción si hay cambio de estado.

  // Lectura inicial del nivel de la batería
  // Voltaje aproximado que cae en el diodo: ~0.7 V
  // Máximo nivel de tensión esperado = 12.6 (batería con carga total) - 0.7 = 11.9 V
  // Divisor de voltaje -> 4K7 + 10K (factor = 0.3197) ---> Máxima tensión en el pin A5 = 3.8 V***  
  // Para hallar voltaje de entrada => lectura*(5/1023)/0.3197 = lectura * 0.0153
  battery_voltage = float(analogRead(BATTERY))*0.0153 + 0.7;
  Serial.print("Tensión inicial: ");
  Serial.print(battery_voltage);
  Serial.println(" V");
  
  // Esperamos hasta que el receptor esté activo y  la señal de potencia esté al mínimo.
  //*
  while(href < 900 || href > 1150 || yawref < 1400){
    href = receiver_input[0];
    yawref = receiver_input[3];
    start ++;                                                               // Mientras esperamos incrementamos la variable "start" en cada ciclo.
    // Enviamos repetidamente una señal mínima a los motores para mantenerlos armados en posición segura
    esc1.writeMicroseconds(1000);
    esc2.writeMicroseconds(1000);
    esc3.writeMicroseconds(1000);
    esc4.writeMicroseconds(1000);    
    delay(3);                                                               
    if (start == 125) {                                                     // Con cada conteo de 125 ciclos...
      digitalWrite(LED_PIN, !digitalRead(LED_PIN));                         // ...cambiamos el estado del LED
      start = 0;                                                            // Reiniciamos "start"
    }
  }
  //*/
  start = 0;                                                                // Reiniciamos "start"
  
  wdt_enable(WDTO_1S);                                                      // Habilitamos watchdog para el loop: 1 segundo    
  led_dis = 1;                                                              // LED arranca en estado encendido
  Serial.println("Listo para volar");
  mtime = millis();                                                         // Inicializamos la variable mtime con el valor actual en milisegundos (para uso del giroscopio)  
  loop_timer = micros();                                                    // Inicializamos la variable mtime con el valor actual en milisegundos (para uso de las interrupciones)  
}


void loop()
{  
  wdt_reset();          // Reset watchdog cada loop

  // Leemos las entradas del transmisor remoto  
  href = receiver_input[0];
  rollref = receiver_input[1];
  pitchref = receiver_input[2];
  yawref = receiver_input[3];
    
  // Leemos la IMU
  read_IMU();
  
  // Obtenemos ángulos de roll and pitch de lecturas del acelerómetro
  rawX = acc[0] - aoffsetX;
  rawY = acc[1] - aoffsetY;
  rawZ = acc[2]  - (4095 - aoffsetZ);         // Debemos considerar la gravedad; en este caso 1g es igual a 4095
  // 1/4096 = 0.0002 (4096 es la sensibilidad según datasheet para el rango que escogimos, usamos el inverso para multiplicar en vez de dividir)
  X = rawX*0.0002;
  Y = rawY*0.0002;
  Z = rawZ*0.0002;
  rollrad = atan(Y/sqrt(X*X+Z*Z));            // Ángulo en radianes
  pitchrad = atan(X/sqrt(Y*Y+Z*Z));           // Ángulo en radianes
  // 180/pi = 57.2958 (factor para convertir a grados)
  angleAx = rollrad*57.2958;                  // Ángulo de roll en grados
  angleAy = pitchrad*57.2958;                 // Ángulo de pitch en grados

  // Código para giroscopio (roll, pitch, yaw)  
  looptime = (millis() - mtime);              // Calculamos el valor transcurrido en ms entre el ciclo pasado y este
  mtime = millis();                           // Guardamos el valor actual para usarlo en el próximo ciclo    
  gyroX = gyro[0] - goffsetX;
  gyroY = gyro[1] - goffsetY;
  gyroZ = gyro[2] - goffsetZ;
  // 1/65.5 = 0.0153 (65.5 es la sensibilidad según datasheet para el rango que escogimos, usamos el inverso para multiplicar en vez de dividir)
  gx_rate = gyroX*0.0153;
  gy_rate = gyroY*0.0153;
  gz_rate = gyroZ*0.0153;
  // Filtrado inicial de las mediciones de velocidad angular usando una proporción de la medida anterior y la actual
  gx_rate = (0.7*last_gx_rate) + (0.3*gx_rate);
  gy_rate = (0.7*last_gy_rate) + (0.3*gy_rate);
  gz_rate = (0.7*last_gz_rate) + (0.3*gz_rate);
  // Integramos para obtener los ángulos
  angleGx = angleGx + (gx_rate * looptime)*0.001;     // Convertir a segundos (1/1000)
  angleGy = angleGy - (gy_rate * looptime)*0.001;     // Medición de pitch rate está invertida por sistema de coordenadas elegido
  angleGz = angleGz - (gz_rate * looptime)*0.001;     // Medición de yaw rate está invertida por sistema de coordenadas elegido
  // Limitamos los ángulos al rango entre -180 y 180 grados
  if (angleGx < -180) angleGx += 360;
  if (angleGx > 180) angleGx -= 360;
  if (angleGy < -180) angleGy += 360;
  if (angleGy > 180) angleGy -= 360;
  if (angleGz < 0) angleGz += 360;
  if (angleGz > 360) angleGz -= 360;
  // Guardar valores actuales para el próximo ciclo
  last_gx_rate = gx_rate;
  last_gy_rate = gy_rate;
  last_gz_rate = gz_rate;

  // Corrección de "yaw" para los ángulos obtenidos del giroscopio
  // pi/180 = 0.0174 (para convertir a radianes, luego *0.001 para convertir a segundos
  angleFx -= angleFy * sin((gz_rate * looptime)*0.0000174);               // Si hubo "yaw" tranferir ángulo de pitch a roll
  angleFy += angleFx * sin((gz_rate * looptime)*0.0000174);               // Si hubo "yaw" tranferir ángulo de roll a pitch 
  // Aplicar el Filtro Complementario a los ángulos originales
  angleFx = 0.995*(angleFx + (gx_rate * looptime*0.001)) + 0.005*angleAx;
  angleFy = 0.995*(angleFy - (gy_rate * looptime*0.001)) + 0.005*angleAy;
  // Limitamos los ángulos al rango entre -180 y 180 grados
  if (angleFx < -180) angleFx += 360;
  if (angleFx > 180) angleFx -= 360;
  if (angleFy < -180) angleFy += 360;
  if (angleFy > 180) angleFy -= 360; 

  // Corrección de velocidad angular para autonivelación
  roll_level_adjust = angleFx * 15;                                      // Calcular corrección de roll
  pitch_level_adjust = angleFy * 15;                                     // Calcular corrección de pitch

  // ------------------------------ ENTRADAS --------------------------------------
  // Para arrancar motores: potencia mínima y yaw a la izquierda (paso 1)
  if (href < 1150 && yawref < 1150) start = 1;
  // Cuando la palanca de yaw retorne al centro , arrancar los motores (paso 2)
  if (start == 1 && href < 1150 && yawref > 1450){
    start = 2;    

    // Reiniciamos algunos parámetros del control PID
    ITerm_r = 0; ITerm_p = 0; ITerm_y = 0;
    lastInput_r = 0; lastInput_p = 0; lastInput_y = 0;
  }
  // Para detener motores: potencia mínima y yaw a la derecha (regreso al paso 0)
  if ((start == 2 && href < 1150 && yawref > 1850) || (href < 900)) start = 0;

  // ------------------------------ CONTROL PID --------------------------------------
  // Roll Control
  // El set point en grados por segundo es determinado de la señal recibida del mando remoto.
  // Dividimos por 3 para limitar el máximo "roll rate" a unos 163 grados por segundo ( (500-10)/3 = 163d/s ).
  Setpoint_r = 0;
  if (href > 1200){                                                       // No hay control si la palanca de potencia está abajo
    // Establecemos una banda muerta de 20us para un mejor rendimiento
    if (rollref > 1510) Setpoint_r = rollref - 1510;
    else if (rollref < 1490) Setpoint_r = rollref - 1490;
  
    Setpoint_r -= roll_level_adjust;                                      // Restar la corrección calculada preciamente para autonivelar
    Setpoint_r *= 0.3333;                                                 // Dividir el setpoint por 3 para obtener ángulos en grados.
  }

  Input_r = gx_rate;                                                      // Como la variable a controlar es la velocidad angular, es este el valor que se pasa como entrada al control PID

  // Pitch Control
  // El set point en grados por segundo es determinado de la señal recibida del mando remoto.
  // Dividimos por 3 para limitar el máximo "pitch rate" a unos 163 grados por segundo ( (500-10)/3 = 163d/s ).
  Setpoint_p = 0;
  if (href > 1200){                                                         // No hay control si la palanca de potencia está abajo
    // Establecemos una banda muerta de 20us para un mejor rendimiento
    if (pitchref > 1510) Setpoint_p = pitchref - 1510;
    else if (pitchref < 1490) Setpoint_p = pitchref - 1490;
  
    Setpoint_p -= pitch_level_adjust;                                       // Restar la corrección calculada preciamente para autonivelar
    Setpoint_p *= 0.3333;                                                   // Dividir el setpoint por 3 para obtener ángulos en grados.
  }

  Input_p = -gy_rate;                                                       // Como la variable a controlar es la velocidad angular, es este el valor que se pasa como entrada al control PID

  // Yaw Control
  // El set point en grados por segundo es determinado de la señal recibida del mando remoto.
  // Dividimos por 3 para limitar el máximo "yaw rate" a unos 163 grados por segundo ( (500-10)/3 = 163d/s ).
  Setpoint_y = 0;
  // Establecemos una banda muerta de 20us para un mejor rendimiento
  if (href > 1200){                                                         // No hay control si la palanca de potencia está abajo
    if (yawref > 1510) Setpoint_y = (yawref - 1510)*0.3333;
    else if (yawref < 1490) Setpoint_y = (yawref - 1490)*0.3333;
  }

  Input_y = gz_rate;                                                        // Como la variable a controlar es la velocidad angular, es este el valor que se pasa como entrada al control PID
  
  // Todos los cálculos de control PID
  Compute();
  // Para señal de potencia se envía el valor recibido directamente del receptor
  Output_h = href;
  
  if (Output_h > 1800) Output_h = 1800;                                     // Dejamos un margen para poder realizar control a máxima potencia

  // Calculamos la salida a cada motor combinando adecuadamente la señal de salida de cada controlador
  int Output1 = Output_h + Output_r + Output_p - Output_y;
  int Output2 = Output_h - Output_r + Output_p + Output_y;
  int Output3 = Output_h - Output_r - Output_p - Output_y;
  int Output4 = Output_h + Output_r - Output_p + Output_y;
  
  // Limitamos los valores de ancho de pulso entre 1000 y 2000 us
  if (Output1 > 2000) Output1 = 2000;
  else if (Output1 < 1000) Output1 = 1000;
  if (Output2 > 2000) Output2 = 2000;
  else if (Output2 < 1000) Output2 = 1000;
  if (Output3 > 2000) Output3 = 2000;
  else if (Output3 < 1000) Output3 = 1000;
  if (Output4 > 2000) Output4 = 2000;
  else if (Output4 < 1000) Output4 = 1000;  
      
  // Verificamos si estamos en el paso 2 (motores operando)
  if (start == 2) {
    // Si la salida es más baja de 1100 microsegundos, mantenemos los motores corriendo a baja potencia
    if (Output1 < 1100) Output1 = 1100;
    if (Output2 < 1100) Output2 = 1100;
    if (Output3 < 1100) Output3 = 1100;
    if (Output4 < 1100) Output4 = 1100;
    
    // Enviamos la respectiva señal a cada motor
    esc1.writeMicroseconds(Output1);
    esc2.writeMicroseconds(Output2);
    esc3.writeMicroseconds(Output3);
    esc4.writeMicroseconds(Output4);

    // El LED se mantiene apagado si el dron está operando
    led_dis = 0;
        
  } else {
    // Si no estamos en el paso 2, no usamos las salidas del sistema de control y encendemos el LED a través de la variable led_dis
    led_dis = 1;
  }

  // Medición del nivel de la batería
  // Usamos un sencillo filtro para reducir el ruido usando una combinación entre la lectura anterior y la nueva    
  battery_voltage = battery_voltage*0.92 + (float(analogRead(BATTERY))*0.0153 + 0.7)*0.08;

  // Encender LED si la batería está baja (menos de 10.3 V).
  if(battery_voltage < 10.3 && battery_voltage > 6-0) led_dis = 2;

  // Rutina para el LED, dependiendo del valor de la variable led_dis
  if (led_dis > 0) {
    digitalWrite(LED_PIN, HIGH);        // Encender LED
    // Motores apagados
    esc1.writeMicroseconds(1000);       
    esc2.writeMicroseconds(1000);
    esc3.writeMicroseconds(1000);
    esc4.writeMicroseconds(1000);
  } else {
    digitalWrite(LED_PIN, LOW);         // Apagar LED
  }  

  // El programa debe correr a 250Hz. Eso significa que debemos asegurar un ciclo cada 4ms.
  while (micros() - loop_timer < 4000);                                       // Esperamos a que transcurran 4000us
  loop_timer = micros();                                                      // Actualizamos el valor del timer para el siguiente ciclo

  // Verificamos estado de comunicación
  signed long timedif;
  timedif = loop_timer - last_int;
  if (timedif > 2000000) {                        // Si han pasado 2 segundo desde la última interrupción, no hay comunicación    
    start = 0;                                   // Mantenemos en estado 0 (dron apagado)
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// RUTINA DE INTERRUPCIÓN
// Esta rutina es invocada cada vez que alguno de los pines 7, 6, 5 o 4 cambia de estado, para leer las señales del receptor.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ISR(PCINT2_vect){
  current_time = micros();
  //Channel 1=========================================
  if(PIND & B10000000){                                                     // ¿El pin 7 está ALTO?
    if(last_channel_1 == 0){                                                // Si cambió de 0 a 1...
      last_channel_1 = 1;                                                   // Guardar estado actual.
      timer_1 = current_time;                                               // Guardar tiempo actual en timer_1.
    }
  }
  else if(last_channel_1 == 1){                                             // Si el pin 7 está en BAJO y pasó de 1 a 0.
    last_channel_1 = 0;                                                     // Guardar estado actual.
    receiver_input[0] = current_time - timer_1;                             // El valor de este canal es current_time - timer_1.
  }
  //Channel 2=========================================
  if(PIND & B01000000 ){                                                    // ¿El pin 6 está ALTO?
    if(last_channel_2 == 0){                                                // Si cambió de 0 a 1...
      last_channel_2 = 1;                                                   // Guardar estado actual.
      timer_2 = current_time;                                               // Guardar tiempo actual en timer_2.
    }
  }
  else if(last_channel_2 == 1){                                             // Si el pin 6 está en BAJO y pasó de 1 a 0.
    last_channel_2 = 0;                                                     // Guardar estado actual.
    receiver_input[1] = current_time - timer_2;                             // El valor de este canal es current_time - timer_2.
  }
  //Channel 3=========================================
  if(PIND & B00100000 ){                                                    // ¿El pin 5 está ALTO?
    if(last_channel_3 == 0){                                                // Si cambió de 0 a 1...
      last_channel_3 = 1;                                                   // Guardar estado actual.
      timer_3 = current_time;                                                // Guardar tiempo actual en timer_3.
    }
  }
  else if(last_channel_3 == 1){                                             // Si el pin 5 está en BAJO y pasó de 1 a 0.
    last_channel_3 = 0;                                                     // Guardar estado actual.
    receiver_input[2] = current_time - timer_3;                             // El valor de este canal es current_time - timer_3.

  }
  //Channel 4=========================================
  if(PIND & B00010000 ){                                                    // ¿El pin 4 está ALTO?
    if(last_channel_4 == 0){                                                // Si cambió de 0 a 1...
      last_channel_4 = 1;                                                   // Guardar estado actual.
      timer_4 = current_time;                                               // Guardar tiempo actual en timer_4.
    }
  }
  else if(last_channel_4 == 1){                                             // Si el pin 4 está en BAJO y pasó de 1 a 0.
    last_channel_4 = 0;                                                     // Guardar estado actual.
    receiver_input[3] = current_time - timer_4;                             // El valor de este canal es current_time - timer_4.
  }
  last_int = current_time;                                                  // Guardamos el tiempo de la última interrupción de cualquier pin
}


// ------------Funciones complementarias---------------

// Inicializar la IMU
void init_IMU() {  
  Wire.beginTransmission(IMU_ADDRESS);                                       // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x6B);                                                          // Escribir al registro PWR_MGMT_1 (6B hex)
  Wire.write(0x00);                                                          // Escribir 00000000 para activar el giroscopio
  Wire.endTransmission();                                                    // Finalizar la transmisión

  Wire.beginTransmission(IMU_ADDRESS);                                       // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x1B);                                                          // Escribir al registro GYRO_CONFIG (1B hex)
  Wire.write(0x08);                                                          // Escribir 00001000 (Rango 500dps)
  Wire.endTransmission();                                                    // Finalizar la transmisión

  Wire.beginTransmission(IMU_ADDRESS);                                       // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x1C);                                                          // Escribir al registro ACCEL_CONFIG (1A hex)
  Wire.write(0x10);                                                          // Escribir 00010000 (Rango +/- 8g)
  Wire.endTransmission();                                                    // Finalizar la transmisión

  // Pequeña revisión para verificar los valores enviados
  Wire.beginTransmission(IMU_ADDRESS);                                       // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x1B);                                                          // Leer registro 0x1B
  Wire.endTransmission();                                                    // Finalizar la transmisión
  Wire.requestFrom(IMU_ADDRESS, 1);                                          // Solicitar 1 byte
  while(Wire.available() < 1);                                               // Esperar la recepción de la información
  if(Wire.read() != 0x08){                                                   // Verificar que el valor es 0x08
    digitalWrite(LED_PIN,HIGH);                                              // Si no, encender LED de alerta
    while(1) delay(10);                                                      // Quedarse en este "loop"
  }

  Wire.beginTransmission(IMU_ADDRESS);                                       // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x1A);                                                          // Escribir al registro CONFIG (1A hex)
  Wire.write(0x03);                                                          // Escribir 00000011 (Low Pass Filter to ~43Hz)
  Wire.endTransmission();                                                    // Finalizar la transmisión
}

// Leer la IMU
void read_IMU(){  
  Wire.beginTransmission(IMU_ADDRESS);                                    // Comenzar comunicación con la dirección de la IMU.
  Wire.write(0x3B);                                                       // Leer desde registro 3B
  Wire.endTransmission();                                                 // Finalizar la transmisión
  Wire.requestFrom(IMU_ADDRESS,14);                                       // Solicitar 14 bytes
      
  while(Wire.available() < 14);                                           // Esperar recepción de la información
  acc[0] = Wire.read()<<8|Wire.read();                                    // Concatenar low y high byte de acelerómetro en eje x.
  acc[1] = Wire.read()<<8|Wire.read();                                    // Concatenar low y high byte de acelerómetro en eje y.
  acc[2] = Wire.read()<<8|Wire.read();                                    // Concatenar low y high byte de acelerómetro en eje z.
  temperature = Wire.read()<<8|Wire.read();                               // Concatenar low y high byte en la variable temperature.
  gyro[0] = Wire.read()<<8|Wire.read();                                   // Concatenar low y high byte de giroscopio en eje x.
  gyro[1] = Wire.read()<<8|Wire.read();                                   // Concatenar low y high byte de giroscopio en eje y.
  gyro[2] = Wire.read()<<8|Wire.read();                                   // Concatenar low y high byte de giroscopio en eje z.
}

// Rutina de calibración
void calibrate() {
  // Leemos cien datos de la IMU y obtenemos los respectivos promedios
  for (int i = 0; i < 100; i++) {
    read_IMU();
    if (i == 0) {      
      goffsetX = gyro[0];
      goffsetY = gyro[1];
      goffsetZ = gyro[2];
    }
    if (i > 0) {      
      goffsetX = (gyro[0] + goffsetX);
      goffsetY = (gyro[1] + goffsetY);
      goffsetZ = (gyro[2] + goffsetZ);
    }
  }  
  goffsetX = goffsetX*0.01;
  goffsetY = goffsetY*0.01;
  goffsetZ = goffsetZ*0.01;

  // Los valores obtenidos son el "offset" inicial de los giroscopios
  Serial.print("Los offset del GYRO son: ");
  Serial.print(goffsetX);
  Serial.print(", ");
  Serial.print(goffsetY);
  Serial.print(", ");
  Serial.println(goffsetZ);
  
  // El "offset" inicial de los acelerómetros debe ser introducido arriba manualmente, dependiendo de su calibración EN REPOSO
  Serial.print("Los offset del ACC son: ");
  Serial.print(aoffsetX);
  Serial.print(", ");
  Serial.print(aoffsetY);
  Serial.print(", ");
  Serial.println(aoffsetZ);
  
  // Volvemos a leer la IMU para obtener datos promedio del acelerómetro
  float acc_x = 0, acc_y = 0, acc_z = 0;
  for (int i = 0; i < 10; i++) {
    read_IMU();
    acc_x += acc[0];
    acc_y += acc[1];
    acc_z += acc[2];
  }
  // Promedio de 10 lecturas
  acc_x *= 0.1;
  acc_y *= 0.1;
  acc_z *= 0.1;  
  rawX = acc_x - aoffsetX;
  rawY = acc_y - aoffsetY;
  rawZ = acc_z  - (4095 - aoffsetZ);         // Debemos considerar la gravedad; en este caso 1g es igual a 4095
  // 1/4096 = 0.0002 (4096 es la sensibilidad según datasheet para el rango que escogimos, usamos el inverso para multiplicar en vez de dividir)
  X = rawX*0.0002;
  Y = rawY*0.0002;
  Z = rawZ*0.0002;
  rollrad = atan(Y/sqrt(X*X+Z*Z));            // Ángulo en radianes
  pitchrad = atan(X/sqrt(Y*Y+Z*Z));           // Ángulo en radianes
  // 180/pi = 57.2958 (factor para convertir a grados)
  angleAx = rollrad*57.2958;                  // Ángulo de roll en grados
  angleAy = pitchrad*57.2958;                 // Ángulo de pitch en grados
  // Inicializamos los filtros con estos ángulos
  angleGx = angleAx;
  angleGy = angleAy;
  angleFx = angleAx;
  angleFy = angleAy;
    
}


//--------Funciones del control PID------
void Compute() {
    // Calculamos todos los errores
    double error_r = Setpoint_r - Input_r;
    double error_p = Setpoint_p - Input_p;
    double error_y = Setpoint_y - Input_y;
    ITerm_r += (ki_r * error_r);
    if (ITerm_r > outMax_dif) ITerm_r = outMax_dif;
    else if (ITerm_r < outMin_dif) ITerm_r = outMin_dif;
    ITerm_p += (ki_p * error_p);
    if (ITerm_p > outMax_dif) ITerm_p = outMax_dif;
    else if (ITerm_p < outMin_dif) ITerm_p = outMin_dif;
    ITerm_y += (ki_y * error_y);
    if (ITerm_y > outMax_dif) ITerm_y = outMax_dif;
    else if (ITerm_y < outMin_dif) ITerm_y = outMin_dif;    
    
    double dInput_r = (Input_r - lastInput_r);
    double dInput_p = (Input_p - lastInput_p);
    double dInput_y = (Input_y - lastInput_y);
    
    // Calculamos la función de salida del PID.
    Output_r = (kp_r * error_r) + ITerm_r - (kd_r * dInput_r);
    if (Output_r > outMax_dif) Output_r = outMax_dif;
    else if (Output_r < outMin_dif) Output_r = outMin_dif;
    Output_p = (kp_p * error_p) + ITerm_p - (kd_p * dInput_p);
    if (Output_p > outMax_dif) Output_p = outMax_dif;
    else if (Output_p < outMin_dif) Output_p = outMin_dif;
    Output_y = (kp_y * error_y) + ITerm_y - (kd_y * dInput_y);
    if (Output_y > outMax_dif) Output_y = outMax_dif;
    else if (Output_y < outMin_dif) Output_y = outMin_dif;
    
    // Guardamos el valor de algunas variables para el próximo recálculo.
    lastInput_r = Input_r;
    lastInput_p = Input_p;
    lastInput_y = Input_y;
    //lastTime = now;
  //}
}

double *SetTunings (double Kp, double Ki, double Kd) {
  static double tunes[3];
  int i;
  if (Kp<0 || Ki<0|| Kd<0) {    
    tunes[0] = 0;
    tunes[1] = 0;
    tunes[2] = 0;
    return tunes;
  }
  double SampleTimeInSec = ((double)SampleTime)/1000;
  tunes[0] = Kp;
  tunes[1] = Ki * SampleTimeInSec;
  tunes[2] = Kd / SampleTimeInSec;
  return tunes;
}

double *SetOutputLimits(double Min, double Max) {
  static double lims[2];
  if (Min > Max) {
    lims[0] = Max;
    lims[1] = Min;
  } else {
    lims[0] = Min;
    lims[1] = Max;
  }
  return lims;
}
